@@ -17,6 +17,7 @@
  * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
  *
  * (C) Copyright 2008 - 2011 Red Hat, Inc.
+ * (C) Copyright 2015,2017,2018 Lubomir Rintel
  */
 
 #include "nm-default.h"
@@ -110,30 +111,50 @@ get_secrets (const char *vpn_uuid,
              gboolean retry,
              gboolean allow_interaction,
              gboolean external_ui_mode,
+             const char *hints[],
              const char *in_pw,
-             char **out_pw,
-             NMSettingSecretFlags pw_flags)
+             char **out_password,
+             char **out_otp,
+             NMSettingSecretFlags password_flags,
+             NMSettingSecretFlags otp_flags)
 {
 	NMAVpnPasswordDialog *dialog;
 	char *prompt, *pw = NULL;
 	const char *new_password = NULL;
+	gboolean ask_password = FALSE;
+	gboolean ask_otp = FALSE;
+	int i;
 
 	g_return_val_if_fail (vpn_uuid != NULL, FALSE);
 	g_return_val_if_fail (vpn_name != NULL, FALSE);
-	g_return_val_if_fail (out_pw != NULL, FALSE);
-	g_return_val_if_fail (*out_pw == NULL, FALSE);
+	g_return_val_if_fail (out_password != NULL, FALSE);
+	g_return_val_if_fail (*out_password == NULL, FALSE);
+	g_return_val_if_fail (out_otp != NULL, FALSE);
+	g_return_val_if_fail (*out_otp == NULL, FALSE);
 
-	/* Get the existing secret, if any */
-	if (   !(pw_flags & NM_SETTING_SECRET_FLAG_NOT_SAVED)
-	    && !(pw_flags & NM_SETTING_SECRET_FLAG_NOT_REQUIRED)) {
+	/* Get the existing password, if any */
+	if (   !(password_flags & NM_SETTING_SECRET_FLAG_NOT_SAVED)
+	    && !(password_flags & NM_SETTING_SECRET_FLAG_NOT_REQUIRED)) {
 		if (in_pw)
 			pw = g_strdup (in_pw);
 		else
 			pw = keyring_lookup_secret (vpn_uuid, NM_FORTISSLVPN_KEY_PASSWORD);
 	}
 
-	/* Don't ask if the passwords is unused */
-	if (pw_flags & NM_SETTING_SECRET_FLAG_NOT_REQUIRED) {
+	if (hints) {
+		for (i = 0; hints[i]; i++) {
+			if (strcmp (hints[i], "password") == 0)
+				ask_password = TRUE;
+			else if (strcmp (hints[i], "otp") == 0)
+				ask_otp = TRUE;
+		}
+	} else {
+		ask_password = !(password_flags & NM_SETTING_SECRET_FLAG_NOT_REQUIRED);
+		ask_otp = (otp_flags & NM_SETTING_SECRET_FLAG_NOT_SAVED);
+	}
+
+	/* Don't ask if the no secrets are needed is unused */
+	if (!ask_password && !ask_otp) {
 		g_free (pw);
 		return TRUE;
 	}
@@ -152,39 +173,51 @@ get_secrets (const char *vpn_uuid,
 		g_key_file_set_string (keyfile, UI_KEYFILE_GROUP, "Description", prompt);
 		g_key_file_set_string (keyfile, UI_KEYFILE_GROUP, "Title", _("Authenticate VPN"));
 
-		keyfile_add_entry_info (keyfile, NM_FORTISSLVPN_KEY_PASSWORD, pw ? pw : "", _("Password:"), TRUE, allow_interaction);
+		if (ask_password)
+			keyfile_add_entry_info (keyfile, NM_FORTISSLVPN_KEY_PASSWORD, pw ? pw : "", _("Password:"), TRUE, allow_interaction);
+		if (ask_otp)
+			keyfile_add_entry_info (keyfile, NM_FORTISSLVPN_KEY_OTP, "", _("Token:"), TRUE, allow_interaction);
 
 		keyfile_print_stdout (keyfile);
 		g_key_file_unref (keyfile);
 		goto out;
 	} else if (   allow_interaction == FALSE
-	           || (!retry && pw && !(pw_flags & NM_SETTING_SECRET_FLAG_NOT_SAVED))) {
+	           || (!retry && pw && !(password_flags & NM_SETTING_SECRET_FLAG_NOT_SAVED) && !ask_otp)) {
 		/* If interaction isn't allowed, just return existing secrets.
 		 * Also, don't ask the user if we don't need a new password (ie, !retry),
 		 * we have an existing PW, and the password is saved.
 		 */
 
-		*out_pw = pw;
+		*out_password = pw;
 		g_free (prompt);
 		return TRUE;
 	}
 
+	gtk_init (NULL, NULL);
 
 	dialog = (NMAVpnPasswordDialog *) nma_vpn_password_dialog_new (_("Authenticate VPN"), prompt, NULL);
 
-	nma_vpn_password_dialog_set_show_password_secondary (dialog, FALSE);
+	/* The one-time-password. */
+	if (ask_otp) {
+		nma_vpn_password_dialog_set_password_secondary_label (dialog, _("_Token:"));
+		nma_vpn_password_dialog_set_show_password_secondary (dialog, TRUE);
+	} else {
+		nma_vpn_password_dialog_set_show_password_secondary (dialog, FALSE);
+	}
 
 	/* pre-fill dialog with the password */
-	if (pw && !(pw_flags & NM_SETTING_SECRET_FLAG_NOT_SAVED))
+	if (pw && !(password_flags & NM_SETTING_SECRET_FLAG_NOT_SAVED))
 		nma_vpn_password_dialog_set_password (dialog, pw);
 
 	gtk_widget_show (GTK_WIDGET (dialog));
 
 	if (nma_vpn_password_dialog_run_and_block (dialog)) {
-
 		new_password = nma_vpn_password_dialog_get_password (dialog);
 		if (new_password)
-			*out_pw = g_strdup (new_password);
+			*out_password = g_strdup (new_password);
+		new_password = nma_vpn_password_dialog_get_password_secondary (dialog);
+		if (new_password)
+			*out_otp = g_strdup (new_password);
 	}
 
 	gtk_widget_hide (GTK_WIDGET (dialog));
@@ -225,9 +258,15 @@ int
 main (int argc, char *argv[])
 {
 	gboolean retry = FALSE, allow_interaction = FALSE, external_ui_mode = FALSE;
-	char *vpn_name = NULL, *vpn_uuid = NULL, *vpn_service = NULL, *password = NULL;
+	gs_free char *vpn_name = NULL;
+	gs_free char *vpn_uuid = NULL;
+	gs_free char *vpn_service = NULL;
+	gs_strfreev char **hints = NULL;
+	char *password = NULL;
+	char *otp = NULL;
 	GHashTable *data = NULL, *secrets = NULL;
-	NMSettingSecretFlags pw_flags = NM_SETTING_SECRET_FLAG_NONE;
+	NMSettingSecretFlags password_flags = NM_SETTING_SECRET_FLAG_NONE;
+	NMSettingSecretFlags otp_flags = NM_SETTING_SECRET_FLAG_NONE;
 	GOptionContext *context;
 	GOptionEntry entries[] = {
 			{ "reprompt", 'r', 0, G_OPTION_ARG_NONE, &retry, "Reprompt for passwords", NULL},
@@ -236,6 +275,7 @@ main (int argc, char *argv[])
 			{ "service", 's', 0, G_OPTION_ARG_STRING, &vpn_service, "VPN service type", NULL},
 			{ "allow-interaction", 'i', 0, G_OPTION_ARG_NONE, &allow_interaction, "Allow user interaction", NULL},
 			{ "external-ui-mode", 0, 0, G_OPTION_ARG_NONE, &external_ui_mode, "External UI mode", NULL},
+			{ "hint", 't', 0, G_OPTION_ARG_STRING_ARRAY, &hints, "Hints from the VPN plugin", NULL},
 			{ NULL }
 		};
 
@@ -243,10 +283,9 @@ main (int argc, char *argv[])
 	bind_textdomain_codeset (GETTEXT_PACKAGE, "UTF-8");
 	textdomain (GETTEXT_PACKAGE);
 
-	gtk_init (&argc, &argv);
-
 	context = g_option_context_new ("- fortisslvpn auth dialog");
 	g_option_context_add_main_entries (context, entries, GETTEXT_PACKAGE);
+	g_option_context_add_group (context, gtk_get_option_group (FALSE));
 	g_option_context_parse (context, &argc, &argv, NULL);
 	g_option_context_free (context);
 
@@ -266,21 +305,28 @@ main (int argc, char *argv[])
 		return 1;
 	}
 
-	nm_vpn_service_plugin_get_secret_flags (secrets, NM_FORTISSLVPN_KEY_PASSWORD, &pw_flags);
+	nm_vpn_service_plugin_get_secret_flags (data, NM_FORTISSLVPN_KEY_PASSWORD, &password_flags);
+	nm_vpn_service_plugin_get_secret_flags (data, NM_FORTISSLVPN_KEY_OTP, &otp_flags);
 
-	if (!get_secrets (vpn_uuid, vpn_name, retry, allow_interaction, external_ui_mode,
+	if (!get_secrets (vpn_uuid, vpn_name, retry, allow_interaction,
+	                  external_ui_mode, (const char **) hints,
 	                  g_hash_table_lookup (secrets, NM_FORTISSLVPN_KEY_PASSWORD),
 	                  &password,
-	                  pw_flags))
+	                  &otp,
+	                  password_flags,
+	                  otp_flags))
 		return 1;
 
 	if (!external_ui_mode) {
 		/* dump the passwords to stdout */
 		if (password)
 			printf ("%s\n%s\n", NM_FORTISSLVPN_KEY_PASSWORD, password);
+		if (otp)
+			printf ("%s\n%s\n", NM_FORTISSLVPN_KEY_OTP, otp);
 		printf ("\n\n");
 
 		g_free (password);
+		g_free (otp);
 
 		/* for good measure, flush stdout since Kansas is going Bye-Bye */
 		fflush (stdout);
