@@ -18,12 +18,12 @@
  * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
  *
  * (C) Copyright 2008 - 2014 Red Hat, Inc.
- * (C) Copyright 2015 Lubomir Rintel
+ * (C) Copyright 2015,2017 Lubomir Rintel
  */
 
-#ifdef HAVE_CONFIG_H
-#include <config.h>
-#endif
+#include "nm-default.h"
+
+#include "nm-fortisslvpn-service.h"
 
 #include <stdio.h>
 #include <string.h>
@@ -37,19 +37,32 @@
 #include <locale.h>
 #include <errno.h>
 
-#include <glib.h>
 #include <glib/gstdio.h>
-#include <glib/gi18n.h>
 
-#include "nm-fortisslvpn-service.h"
+#include "nm-fortissl-properties.h"
 #include "nm-ppp-status.h"
 #include "nm-fortisslvpn-pppd-service-dbus.h"
+#include "nm-utils/nm-shared-utils.h"
+#include "nm-utils/nm-vpn-plugin-macros.h"
 
 #if !defined(DIST_VERSION)
 # define DIST_VERSION VERSION
 #endif
 
-static gboolean debug = FALSE;
+/* openfortissl has configurable log-levels via "-v" command line switch.
+ * See levels in "src/log.h" */
+#define OPENFORTISSL_LOG_MUTE           NULL
+#define OPENFORTISSL_LOG_ERROR          "-v"
+#define OPENFORTISSL_LOG_WARN           "-vv"
+#define OPENFORTISSL_LOG_INFO           "-vvv"
+#define OPENFORTISSL_LOG_DEBUG          "-vvvv"
+#define OPENFORTISSL_LOG_DEBUG_PACKETS  "-vvvvv"
+
+static struct {
+	gboolean debug;
+	int log_level;
+	const char *openfortissl_log_level;
+} gl/*obal*/;
 
 /********************************************************/
 /* The VPN plugin service                               */
@@ -62,7 +75,8 @@ G_DEFINE_TYPE_WITH_CODE (NMFortisslvpnPlugin, nm_fortisslvpn_plugin, NM_TYPE_VPN
 
 typedef struct {
 	GPid pid;
-	guint32 ppp_timeout_handler;
+	gboolean interactive;
+	GDBusMethodInvocation *get_pin_invocation;
 	NMConnection *connection;
 	char *config_file;
 	NMDBusFortisslvpnPpp *dbus_skeleton;
@@ -70,217 +84,29 @@ typedef struct {
 
 #define NM_FORTISSLVPN_PLUGIN_GET_PRIVATE(o) (G_TYPE_INSTANCE_GET_PRIVATE ((o), NM_TYPE_FORTISSLVPN_PLUGIN, NMFortisslvpnPluginPrivate))
 
-#define NM_FORTISSLVPN_PPPD_PLUGIN PLUGINDIR "/nm-fortisslvpn-pppd-plugin.so"
 #define NM_FORTISSLVPN_WAIT_PPPD 10000 /* 10 seconds */
-#define FORTISSLVPN_SERVICE_SECRET_TRIES "fortisslvpn-service-secret-tries"
 
-typedef struct {
-	const char *name;
-	GType type;
-	gboolean required;
-} ValidProperty;
-
-static ValidProperty valid_properties[] = {
-	{ NM_FORTISSLVPN_KEY_GATEWAY,           G_TYPE_STRING, TRUE },
-	{ NM_FORTISSLVPN_KEY_USER,              G_TYPE_STRING, TRUE },
-	{ NM_FORTISSLVPN_KEY_CA,                G_TYPE_STRING, FALSE },
-	{ NM_FORTISSLVPN_KEY_TRUSTED_CERT,      G_TYPE_STRING, FALSE },
-	{ NM_FORTISSLVPN_KEY_CERT,              G_TYPE_STRING, FALSE },
-	{ NM_FORTISSLVPN_KEY_KEY,               G_TYPE_STRING, FALSE },
-	{ NM_FORTISSLVPN_KEY_PASSWORD"-flags",  G_TYPE_UINT, FALSE },
-	{ NULL,                                 G_TYPE_NONE, FALSE }
-};
-
-static ValidProperty valid_secrets[] = {
-	{ NM_FORTISSLVPN_KEY_PASSWORD,          G_TYPE_STRING, TRUE },
-	{ NULL,                                 G_TYPE_NONE,   FALSE }
-};
+#define _NMLOG(level, ...) \
+    G_STMT_START { \
+         if (gl.log_level >= (level)) { \
+              g_printerr ("nm-fortisslvpn[%ld] %-7s " _NM_UTILS_MACRO_FIRST (__VA_ARGS__) "\n", \
+                         (long) getpid (), \
+                          nm_utils_syslog_to_str (level) \
+                          _NM_UTILS_MACRO_REST (__VA_ARGS__)); \
+         } \
+    } G_STMT_END
 
 static gboolean
-validate_gateway (const char *gateway)
+_LOGD_enabled (void)
 {
-	if (!gateway || !strlen (gateway) || !isalnum (*gateway))
-		return FALSE;
-
-	return TRUE;
+	return gl.log_level >= LOG_INFO;
 }
 
-/* This is a bit half-assed. We should check that the user doesn't
- * abuse this to access files he ordinarily shouldn't, but we an't do
- * any better than this for we don't have any information about the
- * identity of the user that activates the connection.
- * We should probably get the certificate inline or something. */
-static gboolean
-validate_ca (const char *ca)
-{
-	struct stat ca_stat;
+#define _LOGD(...) _NMLOG(LOG_INFO,    __VA_ARGS__)
+#define _LOGI(...) _NMLOG(LOG_NOTICE,  __VA_ARGS__)
+#define _LOGW(...) _NMLOG(LOG_WARNING, __VA_ARGS__)
 
-	/* Tolerate only absolute paths */
-	if (!ca || !strlen (ca) || *ca != '/')
-		return FALSE;
-
-	if (stat (ca, &ca_stat) == -1)
-		return FALSE;
-
-	/* Allow only ordinary files */
-	if (!(ca_stat.st_mode & S_IFREG))
-		return FALSE;
-
-	/* Allow only world-readable files */
-	if ((ca_stat.st_mode & 0444) != 0444)
-		return FALSE;
-
-	return TRUE;
-}
-
-typedef struct ValidateInfo {
-	ValidProperty *table;
-	GError **error;
-	gboolean have_items;
-} ValidateInfo;
-
-static void
-validate_one_property (const char *key, const char *value, gpointer user_data)
-{
-	ValidateInfo *info = (ValidateInfo *) user_data;
-	int i;
-
-	if (*(info->error))
-		return;
-
-	info->have_items = TRUE;
-
-	/* 'name' is the setting name; always allowed but unused */
-	if (!strcmp (key, NM_SETTING_NAME))
-		return;
-
-	for (i = 0; info->table[i].name; i++) {
-		ValidProperty prop = info->table[i];
-		long int tmp;
-
-		if (strcmp (prop.name, key))
-			continue;
-
-		switch (prop.type) {
-		case G_TYPE_STRING:
-			if (   !strcmp (prop.name, NM_FORTISSLVPN_KEY_GATEWAY)
-			    && !validate_gateway (value)) {
-				g_set_error (info->error,
-				             NM_VPN_PLUGIN_ERROR,
-				             NM_VPN_PLUGIN_ERROR_BAD_ARGUMENTS,
-				             _("invalid gateway '%s'"),
-				             value);
-				return;
-			} else if (   !strcmp (prop.name, NM_FORTISSLVPN_KEY_CA)
-			           && !validate_ca (value)) {
-				g_set_error (info->error,
-				             NM_VPN_PLUGIN_ERROR,
-				             NM_VPN_PLUGIN_ERROR_BAD_ARGUMENTS,
-				             _("invalid certificate authority '%s'"),
-				             value);
-				return;
-			}
-			return; /* valid */
-		case G_TYPE_UINT:
-			errno = 0;
-			tmp = strtol (value, NULL, 10);
-			if (errno == 0)
-				return; /* valid */
-
-			g_set_error (info->error,
-			             NM_VPN_PLUGIN_ERROR,
-			             NM_VPN_PLUGIN_ERROR_BAD_ARGUMENTS,
-			             _("invalid integer property '%s'"),
-			             key);
-			break;
-		case G_TYPE_BOOLEAN:
-			if (!strcmp (value, "yes") || !strcmp (value, "no"))
-				return; /* valid */
-
-			g_set_error (info->error,
-			             NM_VPN_PLUGIN_ERROR,
-			             NM_VPN_PLUGIN_ERROR_BAD_ARGUMENTS,
-			             _("invalid boolean property '%s' (not yes or no)"),
-			             key);
-			break;
-		default:
-			g_set_error (info->error,
-			             NM_VPN_PLUGIN_ERROR,
-			             NM_VPN_PLUGIN_ERROR_BAD_ARGUMENTS,
-			             _("unhandled property '%s' type %s"),
-			             key, g_type_name (prop.type));
-			break;
-		}
-	}
-
-	/* Did not find the property from valid_properties or the type did not match */
-	if (!info->table[i].name) {
-		g_set_error (info->error,
-		             NM_VPN_PLUGIN_ERROR,
-		             NM_VPN_PLUGIN_ERROR_BAD_ARGUMENTS,
-		             _("property '%s' invalid or not supported"),
-		             key);
-	}
-}
-
-static gboolean
-validate_properties (NMSettingVpn *s_vpn, GError **error)
-{
-	ValidateInfo info = { &valid_properties[0], error, FALSE };
-	int i;
-
-	nm_setting_vpn_foreach_data_item (s_vpn, validate_one_property, &info);
-	if (!info.have_items) {
-		g_set_error (error,
-		             NM_VPN_PLUGIN_ERROR,
-		             NM_VPN_PLUGIN_ERROR_BAD_ARGUMENTS,
-		             "%s",
-		             _("No VPN configuration options."));
-		return FALSE;
-	}
-
-	if (*error)
-		return FALSE;
-
-	/* Ensure required properties exist */
-	for (i = 0; valid_properties[i].name; i++) {
-		ValidProperty prop = valid_properties[i];
-		const char *value;
-
-		if (!prop.required)
-			continue;
-
-		value = nm_setting_vpn_get_data_item (s_vpn, prop.name);
-		if (!value || !strlen (value)) {
-			g_set_error (error,
-			             NM_VPN_PLUGIN_ERROR,
-			             NM_VPN_PLUGIN_ERROR_BAD_ARGUMENTS,
-			             _("Missing required option '%s'."),
-			             prop.name);
-			return FALSE;
-		}
-	}
-
-	return TRUE;
-}
-
-static gboolean
-validate_secrets (NMSettingVpn *s_vpn, GError **error)
-{
-	ValidateInfo info = { &valid_secrets[0], error, FALSE };
-
-	nm_setting_vpn_foreach_secret (s_vpn, validate_one_property, &info);
-	if (!info.have_items) {
-		g_set_error (error,
-		             NM_VPN_PLUGIN_ERROR,
-		             NM_VPN_PLUGIN_ERROR_BAD_ARGUMENTS,
-		             "%s",
-		             _("No VPN secrets!"));
-		return FALSE;
-	}
-
-	return *error ? FALSE : TRUE;
-}
+/*****************************************************************************/
 
 static gboolean
 ensure_killed (gpointer data)
@@ -298,13 +124,23 @@ cleanup_plugin (NMFortisslvpnPlugin *plugin)
 {
 	NMFortisslvpnPluginPrivate *priv = NM_FORTISSLVPN_PLUGIN_GET_PRIVATE (plugin);
 
+	priv->interactive = FALSE;
+
+	if (priv->get_pin_invocation) {
+		g_dbus_method_invocation_return_error_literal (priv->get_pin_invocation,
+		                                               NMV_EDITOR_PLUGIN_ERROR,
+		                                               NMV_EDITOR_PLUGIN_ERROR_FAILED,
+		                                               "Disconnected");
+		priv->get_pin_invocation = NULL;
+	}
+
 	if (priv->pid) {
 		if (kill (priv->pid, SIGTERM) == 0)
 			g_timeout_add (2000, ensure_killed, GINT_TO_POINTER (priv->pid));
 		else
 			kill (priv->pid, SIGKILL);
 
-		g_message ("Terminated ppp daemon with PID %d.", priv->pid);
+		_LOGI ("Terminated ppp daemon with PID %d.", priv->pid);
 		priv->pid = 0;
 	}
 
@@ -325,11 +161,11 @@ openfortivpn_watch_cb (GPid pid, gint status, gpointer user_data)
 	if (WIFEXITED (status))
 		error = WEXITSTATUS (status);
 	else if (WIFSTOPPED (status))
-		g_warning ("openfortivpn stopped unexpectedly with signal %d", WSTOPSIG (status));
+		_LOGW ("openfortivpn stopped unexpectedly with signal %d", WSTOPSIG (status));
 	else if (WIFSIGNALED (status))
-		g_warning ("openfortivpn died with signal %d", WTERMSIG (status));
+		_LOGW ("openfortivpn died with signal %d", WTERMSIG (status));
 	else
-		g_warning ("openfortivpn died from an unknown cause");
+		_LOGW ("openfortivpn died from an unknown cause");
 
 	/* Reap child if needed. */
 	waitpid (priv->pid, NULL, WNOHANG);
@@ -365,17 +201,6 @@ nm_find_openfortivpn (void)
 	return *openfortivpn_binary;
 }
 
-static gboolean
-pppd_timed_out (gpointer user_data)
-{
-	NMFortisslvpnPlugin *plugin = NM_FORTISSLVPN_PLUGIN (user_data);
-
-	g_warning ("Looks like pppd didn't initialize our dbus module");
-	nm_vpn_service_plugin_failure (NM_VPN_SERVICE_PLUGIN (plugin), NM_VPN_CONNECTION_STATE_REASON_SERVICE_START_TIMEOUT);
-
-	return FALSE;
-}
-
 static gboolean
 run_openfortivpn (NMFortisslvpnPlugin *plugin, NMSettingVpn *s_vpn, GError **error)
 {
@@ -384,6 +209,7 @@ run_openfortivpn (NMFortisslvpnPlugin *plugin, NMSettingVpn *s_vpn, GError **err
 	const char *openfortivpn;
 	GPtrArray *argv;
 	const char *value;
+	gs_free char *str_tmp = NULL;
 
 	openfortivpn = nm_find_openfortivpn ();
 	if (!openfortivpn) {
@@ -407,8 +233,8 @@ run_openfortivpn (NMFortisslvpnPlugin *plugin, NMSettingVpn *s_vpn, GError **err
 	value = nm_setting_vpn_get_data_item (s_vpn, NM_FORTISSLVPN_KEY_GATEWAY);
 	g_ptr_array_add (argv, (gpointer) g_strdup (value));
 
-	if (debug)
-		g_ptr_array_add (argv, (gpointer) g_strdup ("-vvv"));
+	if (gl.openfortissl_log_level)
+		g_ptr_array_add (argv, (gpointer) g_strdup (gl.openfortissl_log_level));
 
 	value = nm_setting_vpn_get_data_item (s_vpn, NM_FORTISSLVPN_KEY_CA);
 	if (value) {
@@ -434,11 +260,22 @@ run_openfortivpn (NMFortisslvpnPlugin *plugin, NMSettingVpn *s_vpn, GError **err
 		g_ptr_array_add (argv, (gpointer) g_strdup (value));
 	}
 
+	g_ptr_array_add (argv, (gpointer) g_strdup ("--pinentry"));
+	g_ptr_array_add (argv, (gpointer) g_strdup (LIBEXECDIR "/nm-fortisslvpn-pinentry"));
+
 	g_ptr_array_add (argv, (gpointer) g_strdup ("--pppd-plugin"));
-	g_ptr_array_add (argv, (gpointer) g_strdup (NM_FORTISSLVPN_PPPD_PLUGIN));
+	g_ptr_array_add (argv, (gpointer) g_strdup (PLUGINDIR "/nm-fortisslvpn-pppd-plugin.so"));
+
+	value = nm_setting_vpn_get_data_item (s_vpn, NM_FORTISSLVPN_KEY_REALM);
+	if (value) {
+		g_ptr_array_add (argv, (gpointer) g_strdup ("--realm"));
+		g_ptr_array_add (argv, (gpointer) g_strdup (value));
+	}
 
 	g_ptr_array_add (argv, NULL);
 
+	_LOGD ("start %s", (str_tmp = g_strjoinv (" ", (char **) argv->pdata)));
+
 	if (!g_spawn_async (NULL, (char **) argv->pdata, NULL,
 	                    G_SPAWN_DO_NOT_REAP_CHILD, NULL, NULL, &pid, error)) {
 		g_ptr_array_free (argv, TRUE);
@@ -446,38 +283,24 @@ run_openfortivpn (NMFortisslvpnPlugin *plugin, NMSettingVpn *s_vpn, GError **err
 	}
 	g_ptr_array_free (argv, TRUE);
 
-	g_message ("openfortivpn started with pid %d", pid);
+	_LOGI ("openfortivpn started with pid %d", pid);
 
 	priv->pid = pid;
 	g_child_watch_add (pid, openfortivpn_watch_cb, plugin);
 
-	priv->ppp_timeout_handler = g_timeout_add (NM_FORTISSLVPN_WAIT_PPPD, pppd_timed_out, plugin);
-
 	return TRUE;
 }
 
-static void
-remove_timeout_handler (NMFortisslvpnPlugin *plugin)
-{
-	NMFortisslvpnPluginPrivate *priv = NM_FORTISSLVPN_PLUGIN_GET_PRIVATE (plugin);
-	
-	if (priv->ppp_timeout_handler) {
-		g_source_remove (priv->ppp_timeout_handler);
-		priv->ppp_timeout_handler = 0;
-	}
-}
-
 static gboolean
 handle_set_state (NMDBusFortisslvpnPpp *object,
                   GDBusMethodInvocation *invocation,
                   guint arg_state,
                   gpointer user_data)
 {
-	remove_timeout_handler (NM_FORTISSLVPN_PLUGIN (user_data));
 	if (arg_state == NM_PPP_STATUS_DEAD || arg_state == NM_PPP_STATUS_DISCONNECT)
 		nm_vpn_service_plugin_disconnect (NM_VPN_SERVICE_PLUGIN (user_data), NULL);
 
-	nmdbus_fortisslvpn_ppp_complete_set_state (object, invocation);
+	nmdbus_fortisslvpn_ppp_complete_set_state (object, invocation); /// remove me
 	return TRUE;
 }
 
@@ -487,12 +310,71 @@ handle_set_ip4_config (NMDBusFortisslvpnPpp *object,
                        GVariant *arg_config,
                        gpointer user_data)
 {
-	g_message ("FORTISSLVPN service (IP Config Get) reply received.");
+	_LOGI ("FORTISSLVPN service (IP Config Get) reply received.");
 
-	remove_timeout_handler (NM_FORTISSLVPN_PLUGIN (user_data));
 	nm_vpn_service_plugin_set_ip4_config (NM_VPN_SERVICE_PLUGIN (user_data), arg_config);
 
-	nmdbus_fortisslvpn_ppp_complete_set_ip4_config (object, invocation);
+	nmdbus_fortisslvpn_ppp_complete_set_ip4_config (object, invocation); /// remove me
+	return TRUE;
+}
+
+static gboolean
+handle_get_pin (NMDBusFortisslvpnPpp *object,
+                GDBusMethodInvocation *invocation,
+                const gchar *arg_title,
+                const gchar *arg_desc,
+                const gchar *arg_prompt,
+                const gchar *arg_hint,
+                gpointer user_data)
+{
+	NMFortisslvpnPlugin *plugin = NM_FORTISSLVPN_PLUGIN (user_data);
+	NMFortisslvpnPluginPrivate *priv = NM_FORTISSLVPN_PLUGIN_GET_PRIVATE (plugin);
+	const char *hints[] = { arg_hint, NULL };
+
+	_LOGI ("FORTISSLVPN service (%s) password request received.", arg_hint);
+
+	if (!priv->interactive) {
+		g_dbus_method_invocation_return_error_literal (invocation,
+		                                               NMV_EDITOR_PLUGIN_ERROR,
+		                                               NMV_EDITOR_PLUGIN_ERROR_FAILED,
+		                                               "More secrets required but cannot ask interactively");
+		return TRUE;
+	}
+
+	if (strcmp (arg_hint, NM_FORTISSLVPN_KEY_OTP) == 0) {
+		NMSettingSecretFlags flags = NM_SETTING_SECRET_FLAG_NONE;
+		NMSetting *s_vpn = nm_connection_get_setting (priv->connection, NM_TYPE_SETTING_VPN);
+
+		g_return_val_if_fail (NM_IS_SETTING_VPN (s_vpn), FALSE);
+		nm_setting_get_secret_flags (s_vpn, arg_hint, &flags, NULL);
+		if ((flags & NM_SETTING_SECRET_FLAG_NOT_SAVED) == 0) {
+			g_dbus_method_invocation_return_error (invocation,
+			                                       NMV_EDITOR_PLUGIN_ERROR,
+			                                       NMV_EDITOR_PLUGIN_ERROR_FAILED,
+			                                       "Secret '%s' is not configured as required", arg_hint);
+			return TRUE;
+		}
+	} else if (strcmp (arg_hint, NM_FORTISSLVPN_KEY_PASSWORD) != 0) {
+		/* nm_fortisslvpn_properties_validate_secrets()() is not tolerant
+		 * towards unknown secrets. Don't make NetworkManager add one. */
+		g_dbus_method_invocation_return_error (invocation,
+		                                       NMV_EDITOR_PLUGIN_ERROR,
+		                                       NMV_EDITOR_PLUGIN_ERROR_FAILED,
+		                                       "Secret '%s' is not supported", arg_hint);
+	}
+
+	if (priv->get_pin_invocation) {
+		/* Should not happen! */
+		g_dbus_method_invocation_return_error_literal (priv->get_pin_invocation,
+		                                               NMV_EDITOR_PLUGIN_ERROR,
+		                                               NMV_EDITOR_PLUGIN_ERROR_FAILED,
+		                                               "Superseded by anoher GetPin() call");
+	}
+
+	priv->get_pin_invocation = invocation;
+	nm_vpn_service_plugin_secrets_required (NM_VPN_SERVICE_PLUGIN (plugin),
+	                                        arg_prompt, hints);
+
 	return TRUE;
 }
 
@@ -500,6 +382,7 @@ static gboolean
 get_credentials (NMSettingVpn *s_vpn,
                  const char **username,
                  const char **password,
+                 const char **otp,
                  GError **error)
 {
 	/* Username; try SSLVPN specific username first, then generic username */
@@ -524,6 +407,8 @@ get_credentials (NMSettingVpn *s_vpn,
 		return FALSE;
 	}
 
+	*otp = nm_setting_vpn_get_secret (s_vpn, NM_FORTISSLVPN_KEY_OTP);
+
 	return TRUE;
 }
 
@@ -534,34 +419,43 @@ real_connect (NMVpnServicePlugin *plugin, NMConnection *connection, GError **err
 	NMSettingVpn *s_vpn;
 	mode_t old_umask;
 	gchar *config;
-	const char *username, *password;
+	const char *username, *password, *realm, *otp;
 
 	g_return_val_if_fail (error == NULL || *error == NULL, FALSE);
 
 	s_vpn = NM_SETTING_VPN (nm_connection_get_setting (connection, NM_TYPE_SETTING_VPN));
-	g_assert (s_vpn);
+	g_return_val_if_fail (NM_IS_SETTING_VPN (s_vpn), FALSE);
 
-	if (!validate_properties (s_vpn, error))
+	if (!nm_fortisslvpn_properties_validate (s_vpn, error))
 		return FALSE;
 
-	if (!validate_secrets (s_vpn, error))
+	if (!nm_fortisslvpn_properties_validate_secrets (s_vpn, error))
 		return FALSE;
 
-	if (!get_credentials (s_vpn, &username, &password, error))
+	if (!get_credentials (s_vpn, &username, &password, &otp, error))
 		return FALSE;
 
+	realm = nm_setting_vpn_get_data_item (s_vpn, NM_FORTISSLVPN_KEY_REALM);
+
 	g_clear_object (&priv->connection);
 	priv->connection = g_object_ref (connection);
 
-	if (debug)
+	if (_LOGD_enabled ()) {
+		_LOGD ("connection:");
 		nm_connection_dump (connection);
+	}
 
 	/* Create the configuration file so that we don't expose
 	 * secrets on the command line */
 	priv->config_file = g_strdup_printf (NM_FORTISSLVPN_STATEDIR "/%s.config",
 	                                     nm_connection_get_uuid (connection));
-	config = g_strdup_printf ("username = %s\npassword = %s\n",
-	                          username, password);
+	config = g_strdup_printf ("username = %s\n"
+	                          "password = %s"
+	                          "%s%s"
+	                          "%s%s\n",
+	                          username, password,
+	                          realm ? "\nrealm = " : "", realm ? realm : "",
+	                          otp ? "\notp = " : "", otp ? otp : "");
 	old_umask = umask (0077);
 	if (!g_file_set_contents (priv->config_file, config, -1, error)) {
 		g_clear_pointer (&priv->config_file, g_free);
@@ -576,6 +470,15 @@ real_connect (NMVpnServicePlugin *plugin, NMConnection *connection, GError **err
 	return run_openfortivpn (NM_FORTISSLVPN_PLUGIN (plugin), s_vpn, error);
 }
 
+static gboolean
+real_connect_interactive (NMVpnServicePlugin *plugin, NMConnection *connection,
+                          GVariant *details, GError **error)
+{
+	NMFortisslvpnPluginPrivate *priv = NM_FORTISSLVPN_PLUGIN_GET_PRIVATE (plugin);
+	priv->interactive = TRUE;
+	return real_connect (plugin, connection, error);
+}
+
 static gboolean
 real_need_secrets (NMVpnServicePlugin *plugin,
                    NMConnection *connection,
@@ -590,19 +493,17 @@ real_need_secrets (NMVpnServicePlugin *plugin,
 
 	s_vpn = nm_connection_get_setting (connection, NM_TYPE_SETTING_VPN);
 
-	nm_setting_get_secret_flags (NM_SETTING (s_vpn), NM_FORTISSLVPN_KEY_PASSWORD, &flags, NULL);
-
-	/* Don't need the password if it's not required */
-	if (flags & NM_SETTING_SECRET_FLAG_NOT_REQUIRED)
-		return FALSE;
-
-	/* Don't need the password if we already have one */
-	if (nm_setting_vpn_get_secret (NM_SETTING_VPN (s_vpn), NM_FORTISSLVPN_KEY_PASSWORD))
-		return FALSE;
-
-	/* Otherwise we need a password */
 	*setting_name = NM_SETTING_VPN_SETTING_NAME;
-	return TRUE;
+
+	/* Do we require the password and don't have it? */
+	nm_setting_get_secret_flags (NM_SETTING (s_vpn), NM_FORTISSLVPN_KEY_PASSWORD, &flags, NULL);
+	if (   !(flags & NM_SETTING_SECRET_FLAG_NOT_REQUIRED)
+	    && !nm_setting_vpn_get_secret (NM_SETTING_VPN (s_vpn), NM_FORTISSLVPN_KEY_PASSWORD))
+		return TRUE;
+
+	/* Otherwise we're fine */
+	*setting_name = NULL;
+	return FALSE;
 }
 
 static gboolean
@@ -612,21 +513,73 @@ real_disconnect (NMVpnServicePlugin *plugin, GError **err)
 	return TRUE;
 }
 
+static gboolean
+real_new_secrets (NMVpnServicePlugin *plugin, NMConnection *connection, GError **error)
+{
+	NMFortisslvpnPluginPrivate *priv = NM_FORTISSLVPN_PLUGIN_GET_PRIVATE (plugin);
+	GVariant *parameters;
+	NMSettingVpn *s_vpn;
+	const char *hint = NULL;
+	const char *pin;
+
+	if (!priv->get_pin_invocation) {
+		g_set_error (error,
+		             NM_VPN_PLUGIN_ERROR,
+		             NM_VPN_PLUGIN_ERROR_LAUNCH_FAILED,
+		             "%s",
+		             _("Got new secrets, but nobody asked for them."));
+		return FALSE;
+	}
+
+	s_vpn = NM_SETTING_VPN (nm_connection_get_setting (connection, NM_TYPE_SETTING_VPN));
+	if (!s_vpn) {
+		g_dbus_method_invocation_return_error_literal (priv->get_pin_invocation,
+		                                               NMV_EDITOR_PLUGIN_ERROR,
+		                                               NMV_EDITOR_PLUGIN_ERROR_FAILED,
+		                                               "Connection lacked VPN setting");
+		goto out;
+	}
+
+	parameters = g_dbus_method_invocation_get_parameters (priv->get_pin_invocation);
+	g_variant_get_child (parameters, 3 /* hint */, "&s", &hint);
+	if (!hint) {
+		/* Can not happen. */
+		g_dbus_method_invocation_return_error_literal (priv->get_pin_invocation,
+		                                               NMV_EDITOR_PLUGIN_ERROR,
+		                                               NMV_EDITOR_PLUGIN_ERROR_FAILED,
+		                                               "Forgotten the secrets hint?");
+		goto out;
+	}
+
+	pin = nm_setting_vpn_get_secret (s_vpn, hint);
+	if (!pin) {
+		g_dbus_method_invocation_return_error (priv->get_pin_invocation,
+		                                       NMV_EDITOR_PLUGIN_ERROR,
+		                                       NMV_EDITOR_PLUGIN_ERROR_FAILED,
+		                                       "No '%s' hint in VPN setting", hint);
+		goto out;
+	}
+
+	nmdbus_fortisslvpn_ppp_complete_get_pin (priv->dbus_skeleton,
+	                                         priv->get_pin_invocation,
+	                                         pin);
+
+out:
+	priv->get_pin_invocation = NULL;
+	return TRUE;
+}
+
 static void
 state_changed_cb (GObject *object, NMVpnServiceState state, gpointer user_data)
 {
 	NMFortisslvpnPluginPrivate *priv = NM_FORTISSLVPN_PLUGIN_GET_PRIVATE (object);
 
 	switch (state) {
-	case NM_VPN_SERVICE_STATE_STARTED:
-		remove_timeout_handler (NM_FORTISSLVPN_PLUGIN (object));
-		break;
 	case NM_VPN_SERVICE_STATE_UNKNOWN:
 	case NM_VPN_SERVICE_STATE_INIT:
 	case NM_VPN_SERVICE_STATE_SHUTDOWN:
 	case NM_VPN_SERVICE_STATE_STOPPING:
 	case NM_VPN_SERVICE_STATE_STOPPED:
-		remove_timeout_handler (NM_FORTISSLVPN_PLUGIN (object));
 		g_clear_object (&priv->connection);
 		break;
 	default:
@@ -650,6 +603,7 @@ dispose (GObject *object)
 			g_dbus_interface_skeleton_unexport (skeleton);
 		g_signal_handlers_disconnect_by_func (skeleton, handle_set_state, object);
 		g_signal_handlers_disconnect_by_func (skeleton, handle_set_ip4_config, object);
+		g_signal_handlers_disconnect_by_func (skeleton, handle_get_pin, object);
 	}
 
 	G_OBJECT_CLASS (nm_fortisslvpn_plugin_parent_class)->dispose (object);
@@ -671,8 +625,10 @@ nm_fortisslvpn_plugin_class_init (NMFortisslvpnPluginClass *fortisslvpn_class)
 	/* virtual methods */
 	object_class->dispose = dispose;
 	parent_class->connect = real_connect;
+	parent_class->connect_interactive = real_connect_interactive;
 	parent_class->need_secrets = real_need_secrets;
 	parent_class->disconnect = real_disconnect;
+	parent_class->new_secrets = real_new_secrets;
 }
 
 static GInitableIface *ginitable_parent_iface = NULL;
@@ -705,6 +661,7 @@ init_sync (GInitable *object, GCancellable *cancellable, GError **error)
 
 	g_signal_connect (priv->dbus_skeleton, "handle-set-state", G_CALLBACK (handle_set_state), object);
 	g_signal_connect (priv->dbus_skeleton, "handle-set-ip4-config", G_CALLBACK (handle_set_ip4_config), object);
+	g_signal_connect (priv->dbus_skeleton, "handle-get-pin", G_CALLBACK (handle_get_pin), object);
 
 	g_object_unref (connection);
 	return TRUE;
@@ -725,10 +682,10 @@ nm_fortisslvpn_plugin_new (const char *bus_name)
 
 	plugin = (NMFortisslvpnPlugin *) g_initable_new (NM_TYPE_FORTISSLVPN_PLUGIN, NULL, &error,
 	                                                 NM_VPN_SERVICE_PLUGIN_DBUS_SERVICE_NAME, bus_name,
-	                                                 NM_VPN_SERVICE_PLUGIN_DBUS_WATCH_PEER, !debug,
+	                                                 NM_VPN_SERVICE_PLUGIN_DBUS_WATCH_PEER, !gl.debug,
 	                                                 NULL);
 	if (!plugin) {
-		g_warning ("Failed to initialize a plugin instance: %s", error->message);
+		_LOGW ("Failed to initialize a plugin instance: %s", error->message);
 		g_error_free (error);
 	}
 
@@ -748,18 +705,19 @@ main (int argc, char *argv[])
 	GMainLoop *main_loop;
 	gboolean persist = FALSE;
 	GOptionContext *opt_ctx = NULL;
-	gchar *bus_name = NM_DBUS_SERVICE_FORTISSLVPN;
+	gs_free char *bus_name_free = NULL;
+	const char *bus_name;
+	GError *error = NULL;
+	char sbuf[30];
 
 	GOptionEntry options[] = {
-		{ "persist", 0, 0, G_OPTION_ARG_NONE, &persist, N_("Don't quit when VPN connection terminates"), NULL },
-		{ "debug", 0, 0, G_OPTION_ARG_NONE, &debug, N_("Enable verbose debug logging (may expose passwords)"), NULL },
-		{ "bus-name", 0, 0, G_OPTION_ARG_STRING, &bus_name, N_("D-Bus name to use for this instance"), NULL },
+		{ "persist", 0, 0, G_OPTION_ARG_NONE, &persist, N_("Don’t quit when VPN connection terminates"), NULL },
+		{ "debug", 0, 0, G_OPTION_ARG_NONE, &gl.debug, N_("Enable verbose debug logging (may expose passwords)"), NULL },
+		{ "bus-name", 0, 0, G_OPTION_ARG_STRING, &bus_name_free, N_("D-Bus name to use for this instance"), NULL },
 		{NULL}
 	};
 
-#if !GLIB_CHECK_VERSION (2, 35, 0)
-	g_type_init ();
-#endif
+	nm_g_type_init ();
 
 	/* locale will be set according to environment LC_* variables */
 	setlocale (LC_ALL, "");
@@ -778,17 +736,45 @@ main (int argc, char *argv[])
 	g_option_context_set_summary (opt_ctx,
 	    _("nm-fortisslvpn-service provides integrated SSLVPN capability (compatible with Fortinet) to NetworkManager."));
 
-	g_option_context_parse (opt_ctx, &argc, &argv, NULL);
+	if (!g_option_context_parse (opt_ctx, &argc, &argv, &error)) {
+		g_printerr ("Error parsing the command line options: %s\n", error->message);
+		g_option_context_free (opt_ctx);
+		g_error_free (error);
+		return EXIT_FAILURE;
+	}
 	g_option_context_free (opt_ctx);
 
+	bus_name = bus_name_free ?: NM_DBUS_SERVICE_FORTISSLVPN;
+
 	if (getenv ("NM_PPP_DEBUG"))
-		debug = TRUE;
+		gl.debug = TRUE;
 
-	if (debug)
-		g_message ("nm-fortisslvpn-service (version " DIST_VERSION ") starting...");
+	gl.log_level = _nm_utils_ascii_str_to_int64 (getenv ("NM_VPN_LOG_LEVEL"),
+	                                             10, 0, LOG_DEBUG, -1);
+	if (gl.log_level < 0)
+		gl.openfortissl_log_level = gl.debug ? OPENFORTISSL_LOG_INFO : OPENFORTISSL_LOG_MUTE;
+	else if (gl.log_level <= 0)
+		gl.openfortissl_log_level = OPENFORTISSL_LOG_MUTE;
+	else if (gl.log_level <= LOG_ERR)
+		gl.openfortissl_log_level = OPENFORTISSL_LOG_ERROR;
+	else if (gl.log_level <= LOG_WARNING)
+		gl.openfortissl_log_level = OPENFORTISSL_LOG_WARN;
+	else if (gl.log_level <= LOG_NOTICE)
+		gl.openfortissl_log_level = OPENFORTISSL_LOG_INFO;
+	else if (gl.log_level <= LOG_INFO)
+		gl.openfortissl_log_level = OPENFORTISSL_LOG_DEBUG;
+	else
+		gl.openfortissl_log_level = OPENFORTISSL_LOG_DEBUG_PACKETS;
 
-	if (bus_name)
-		setenv ("NM_DBUS_SERVICE_FORTISSLVPN", bus_name, 0);
+	if (gl.log_level < 0)
+		gl.log_level = gl.debug ? LOG_DEBUG : LOG_NOTICE;
+
+	_LOGD ("nm-fortisslvpn-service (version " DIST_VERSION ") starting...");
+	_LOGD ("   uses%s --bus-name \"%s\"", bus_name_free ? "" : " default", bus_name);
+
+	setenv ("NM_VPN_LOG_LEVEL", nm_sprintf_buf (sbuf, "%d", gl.log_level), TRUE);
+	setenv ("NM_VPN_LOG_PREFIX_TOKEN", nm_sprintf_buf (sbuf, "%ld", (long) getpid ()), TRUE);
+	setenv ("NM_DBUS_SERVICE_FORTISSLVPN", bus_name, 0);
 
 	plugin = nm_fortisslvpn_plugin_new (bus_name);
 	if (!plugin)
